#define sexp_so_extension ".so"
#define sexp_default_module_path "/usr/local/share/chibi:/usr/local/lib/chibi:/usr/local/share/snow:/usr/local/lib/snow"
#define sexp_platform "bsd"
#define sexp_version "0.10.0"
#define sexp_release_name "neon"
